/*
 * CONNECT.C
 *
 * Functions to provide for a Connect dialog that uses the Net Driver
 * server browse dialog if available as well as functions to manage
 * connections between ourselves and peers.
 *
 * Copyright (c)1992 Microsoft Corporation, All Right Reserved
 *
 * Kraig Brockschmidt, Software Design Engineer
 * Microsoft Systems Developer Relations
 * One Microsoft Way
 * Redmond, WA  98052
 *
 * Internet  :  kraigb@microsoft.com
 * Compuserve:  70750,2344
 * Fax       :  (206)936-7329
 */

#define STRICT
#include <windows.h>
#include <ddeml.h>
#include "exchange.h"
#include "connect.h"





/*
 * GetPeerName
 *
 * Purpose:
 *  Invokes a dialog that asks the user for a network server name
 *  with which we'll connect and exchange data.  The dialog invoked
 *  may exist in the net driver or may be our own if the net driver
 *  does no have one.
 *
 * Parameters:
 *  hWnd            HWND of the parent of the dialog.
 *  hInst           HINSTANCE of the application instance.
 *  pszLocal        char * containing our local name for comparison.
 *  pszName         char * in which to store the chosen name.
 *  cchName         UINT length of pszName.
 *
 * Return Value:
 *  BOOL            TRUE if a name was retrieved, FALSE otherwise.
 */

BOOL GetPeerName(HWND hWnd, HINSTANCE hInst, char *pszLocal, char *pszName, UINT cchName)
    {
    char                szTemp[40];
    LPFNWINNETBROWSE    pfnBrowse;
    HINSTANCE           hModNet;
    FARPROC             pfn;
    BOOL                fRet;
    int                 i;
    NETDLGINFO          ndi;
    WORD                wRet=WN_NOT_SUPPORTED;


    if (!IsWindow(hWnd) || NULL==pszName || 0==cchName)
        return FALSE;

    //This loop allows us to retry everything without using a goto :)
    while (TRUE)
        {
        fRet=FALSE;
        szTemp[0]=0;
        hModNet=(HINSTANCE)WNetGetCaps(0xFFFF);

        if (NULL!=hModNet)
            {
            pfnBrowse=(LPFNWINNETBROWSE)GetProcAddress(hModNet, (LPSTR)(LONG)146);

            if (NULL!=pfnBrowse)
                {
                wRet=(*pfnBrowse)(hWnd, pszLocal, szTemp, sizeof(szTemp), 0L);

                if (WN_CANCEL==wRet)
                    return FALSE;       //User might have cancelled, so leave.

                fRet=(WN_SUCCESS==wRet);
                }
            }

        /*
         * If we tried the other dialog and it failed, or if we can't get to it,
         * use a simpler default dialog.
         */
        if (WN_SUCCESS!=wRet && (NULL==hModNet || NULL==pfnBrowse))
            {
            pfn=MakeProcInstance((FARPROC)ConnectDialogProc, hInst);

            ndi.cch=sizeof(szTemp);
            ndi.psz=szTemp;
            fRet=(BOOL)DialogBoxParam(hInst, (LPSTR)(LONG)IDD_CONNECT
                                      , hWnd, (DLGPROC)pfn, (LONG)(LPSTR)&ndi);
            FreeProcInstance(pfn);
            }


        if (fRet)
            {
            //Strip a \ or \\ from the front of the machine name.
            i=0;
            if ('\\'==szTemp[0])
                i++;

            if ('\\'==szTemp[1])
                i++;

            lstrcpyn(pszName, szTemp+i, cchName);
            }

        //Check if we tried connecting to ourselves.
        if (!lstrcmpi(pszLocal, pszName))
            {
            //Yes, this is a justified use of a goto.
            if (IDRETRY==MessageBox(hWnd, "Connections to yourself are not allowed."
                                    , "Connect...", MB_RETRYCANCEL))
                continue;

            fRet=FALSE;
            }

        break;
        }

    return fRet;
    }




/*
 * ConnectDialogProc
 *
 * Purpose:
 *  Dialog procedure for an cheap application-default-can't-get-anything-
 *  better-from-the-net-driver request.
 *
 * Parameters:
 *  Standard
 *
 * Return Value:
 *  Standard
 */

BOOL FAR PASCAL ConnectDialogProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam)
    {
    static PNETDLGINFO   pNDI;
    WORD                 wID;
    WORD                 wCode;
    HWND                 hWndMsg;


    hWndMsg=(HWND)(UINT)lParam;
    wID=LOWORD(wParam);
    #ifdef WIN32
        wCode=HIWORD(wParam);
    #else
        wCode=HIWORD(lParam);
    #endif


    switch (iMsg)
        {
        case WM_INITDIALOG:
            pNDI=(PNETDLGINFO)(UINT)lParam;
            SendDlgItemMessage(hDlg, ID_SERVERNAME, EM_LIMITTEXT, pNDI->cch, 0L);
            return TRUE;

        case WM_COMMAND:
            switch (wID)
                {
                case IDOK:
                    GetDlgItemText(hDlg, ID_SERVERNAME, pNDI->psz, pNDI->cch);
                    EndDialog(hDlg, TRUE);
                    break;

                case IDCANCEL:
                    EndDialog(hDlg, FALSE);
                    break;

                default:
                    break;
                }
            break;

        default:
            break;
        }

    return FALSE;
    }






/*
 * FEstablishFirstConnection
 *
 * Purpose:
 *  To initially get into the game, you have to connect to anyone
 *  else who's playing.  When you've accomplished that, you can get
 *  from it the list of the others in the game and connect to them
 *  as well.
 *
 * Parameters:
 *  pAV             PAPPVARS to the application variables.
 *
 * Return Value:
 *  BOOL            TRUE if a we connected to DDEML, FALSE otherwise.
 */

BOOL FEstablishFirstConnection(PAPPVARS pAV)
    {
    HCONV       hConv;
    BOOL        fRet=FALSE;

    ODS("FEstablishFirstConnection Entry.");

    /*
     * Connecting to a peer has three parts:
     *  1.  Retrieve a peer computer name.
     *  2.  Call DdeConnect on \\peer\service.  This will only fail if
     *      the local DDEML fails, not if the peer is unavailable.
     *  3.  Start a transaction to confirm that the peer is connected.
     *
     * HConnectToPeer handles 2 & 3.
     */

    //1. Ask the user for a peer computer name. This checks for self-connects
    if (GetPeerName(pAV->hWnd, pAV->hInst, pAV->szLocal, pAV->szPeer, sizeof(pAV->szPeer)))
        {
        hConv=HConnectToPeer(pAV, pAV->szPeer);

        fRet=(NULL!=hConv);

        //Connect to everthing our peer is connected to.  ASYNC Dependent
        if (hConv && MULTI==pAV->iDL)
            fRet=FSyncWithPeer(pAV, hConv);
        }

    D(if (fRet) ODS("FEstablishFirstConnection Confirmed."));
    return fRet;
    }




/*
 * HConnectToPeer
 *
 * Purpose:
 *  Does the actual DDEML calls to connect to a specific peer defined
 *  by pszPeer.  We use DdeConnect then try to poke our local name to
 *  the other machine if necessary.  This function also allocates
 *  a POSITION structure that the XTYP_ADVDATA transaction looks up
 *  to update peer data.
 *
 * Parameters:
 *  pAV             PAPPVARS to the application variables.
 *  pszPeer         LPSTR name of the computer to which we connect.
 *
 * Return Value:
 *  HCONV           Conversation handle if succedful, NULL otherwise.
 */

HCONV HConnectToPeer(PAPPVARS pAV, LPSTR pszPeer)
    {
    HCONV       hConv;
    HSZ         hszMachine;

    ODS("HConnectToPeer Entry.");

    /*
     * Build the \\peer\service name (using our own service since we're
     * trying to connect to ourselves on another computer), store that
     * name in a DDEML HSZ, and call DdeConnect.
     */
    wsprintf(pAV->szTemp, "\\\\%s\\%s", (LPSTR)pszPeer, (LPSTR)PSZ(IDS_NETWORKDDE));
    hszMachine=DdeCreateStringHandle(pAV->dwDDE, pAV->szTemp, CP_WINANSI);

    hConv=DdeConnect(pAV->dwDDE, hszMachine, pAV->hszShare, NULL);
    DdeFreeStringHandle(pAV->dwDDE, hszMachine);

    if (NULL==hConv)
        {
        ODSsz("HConnectToPeer: Connection to %s failed.", (LPSTR)pszPeer);
        return NULL;
        }

    ODSlu("HConnectToPeer DdeConnect hConv=%lu", (DWORD)hConv);

    /*
     * DdeConnect succeeded, but that only means we are talking to the
     * local DDEML/NDDE modules.  We have to try doing something with
     * our peer, so we poke our machine name back to it.
     */

    if (!FPokeLocalName(pAV, pszPeer, hConv))
        {
        DdeDisconnect(hConv);
        hConv=NULL;
        }

    D(if (NULL==hConv) ODS("HConnectToPeer Failed in FPokeLocalName."));
    D(if (NULL!=hConv) ODS("HConnectToPeer Confirmed."));
    return hConv;
    }






/*
 * FPokeLocalName
 *
 * Purpose:
 *  To determine if a Network DDE conversation is actually established,
 *  we have to try something with the peer acting as a server.  In
 *  this case we try to shove our machine name over.  If it works,
 *  then we have a connection.
 *
 * Parameters:
 *  pAV             PAPPVARS to the application variables.
 *  pszPeer         LPSTR for the peer name.
 *  hConv           HCONV of the conversation from DdeConnect.
 *
 * Return Value:
 *  BOOL            TRUE if the function succeeds, FALSE otherwise.
 */

BOOL FPokeLocalName(PAPPVARS pAV, LPSTR pszPeer, HCONV hConv)
    {
    BOOL            fRet=TRUE;
    int             i;

    ODSlu("FPokeLocalName Entry hConv=%lu", (DWORD)hConv);

    i=(int)SendMessage(pAV->hListC, LB_ADDSTRING, 0, (LONG)pszPeer);
    SendMessage(pAV->hListC, LB_SETITEMDATA, i, (DWORD)hConv);

    /*
     * We have to set the XActPending flag to TRUE here before we even
     * start the transaction since Network DDE may be so fast as to send
     * XTYP_XACT_COMPLETE before we even start waiting for it.
     */
    pAV->fXActPending=TRUE;

    //Send our computername to our peer--we loaded it earlier
    DdeClientTransaction((LPVOID)pAV->szLocal, lstrlen(pAV->szLocal)+1, hConv
        , pAV->hszItemLocal, CF_TEXT, XTYP_POKE, (DWORD)TIMEOUT_ASYNC, NULL);

    fRet=FDDETransactionWait(pAV);

    D(i=(int)DdeGetLastError(pAV->dwDDE));
    ODSu("XTYP_POKE Error=%u", i);

    if (fRet)
        {
        //Start advising peer.  This is syncronous
        DdeClientTransaction(NULL, 0L, hConv, pAV->hszItemData, pAV->cfItem
                             , XTYP_ADVSTART, 20000L, NULL);

        i=(int)DdeGetLastError(pAV->dwDDE);
        fRet=(DMLERR_NO_ERROR==i);

        ODSu("XTYP_ADVSTART Error=%u", i);
        }


    if (!fRet)
        {
        SendMessage(pAV->hListC, LB_DELETESTRING, i, 0L);
        ODS("FPokeLocalName Failed.");
        }
    else
        {
        ODS("FPokeLocalName Confirmed.");
        }

    return fRet;
    }





/*
 * FSyncWithPeer
 *
 * Purpose:
 *  Synchronizes our machine with the peer we just connected to.  This
 *  only happens on our initial connection so the conversation is the
 *  first entry in pAV->hListC.  For this we'll send XTYP_REQUEST with the
 *  item name of "ITEM_PEERLIST."  Once we have that list, we can
 *  connect to each peer in turn.
 *
 * Parameters:
 *  pAV             PAPPVARS to the application variables.
 *  hConv           HCONV with which to sync.
 *
 * Return Value:
 *  BOOL            TRUE if the function succeeded, FALSE otherwise.
 */

BOOL FSyncWithPeer(PAPPVARS pAV, HCONV hConv)
    {
    HDDEDATA        hData;
    LPSTR           psz, pszT;
    DWORD           cch;
    char            ch;
    BOOL            fRet=TRUE;

    ODS("FSyncWithPeer Entry.");

    //Poking the local name must have worked, so get the data we want.
    hData=DdeClientTransaction(NULL, 0L, hConv, pAV->hszItemPeer, CF_TEXT
                              , XTYP_REQUEST, 20000L, NULL);

    if (NULL==hData)
        {
        ODS("Request of peer list failed.");
        TerminateConnection(pAV, 0);
        return FALSE;
        }

    pAV->fBlockRequest=TRUE;

    //We got the list, start connecting...
    psz=(LPSTR)DdeAccessData(hData, &cch);
    ch=*psz;

    while (ch!=0 && cch!=0L)
        {
        pszT=psz;

        //Search for the next NULL-term or \ delimeter
        while (0!=*psz && '\\'!=*psz)
            psz++;

        //Save the character we terminated on and make pszT a null-term string.
        ch=*psz;
        *psz++=0;

        //Prevent self-connections
        if (!lstrcmpi(pAV->szLocal, pszT))
            {
            ODSsz("Bypassed list connect to %s", pszT);
            continue;
            }

        //Got a peer name, let's connect!
        if (!HConnectToPeer(pAV, pszT))
            {
            ODSsz("Failed list connect to %s", pszT);
            fRet=FALSE;
            break;
            }

        //Everything else is taken care of already...so continue.
        ODSsz("List connected to %s", pszT);
        }

    //If we failed, we have to get rid of all connections we might have made.
    if (!fRet)
        TerminateAllConnections(pAV);

    DdeUnaccessData(hData);

    pAV->fBlockRequest=FALSE;

    D(if (fRet)  ODS("FSyncWithPeer Confirmed."));
    D(if (!fRet) ODS("FSyncWithPeer Failed."));
    return fRet;
    }






/*
 * TerminateConnection
 *
 * Purpose:
 *  Closes a connection to a single peer.  This function really serves to
 *  isolate disconnection and cleanup code.
 *
 * Parameters:
 *  pAV             PAPPVARS to the application variables.
 *  iConv           int index of the conversation to terminate.
 *
 * Return Value:
 *  None
 */

void TerminateConnection(PAPPVARS pAV, int iConv)
    {
    HCONV           hConv;

   #ifdef DEBUG
    char        szPeer[40];
   #endif

    ODS("TerminateConnection Entry.");

    hConv=(HCONV)SendMessage(pAV->hListC, LB_GETITEMDATA, iConv, 0L);

    //Stop advising first.
    DdeClientTransaction(NULL, 0L, hConv, pAV->hszItemData
                         , pAV->cfItem, XTYP_ADVSTOP, 20000L, NULL);

    D(SendMessage(pAV->hListC, LB_GETTEXT, iConv, (LONG)(LPSTR)szPeer));
    ODSsz("Stopped advising %s", (LPSTR)szPeer);


    //Now end the connection and free the data from that machine.
    if (DdeDisconnect(hConv))
        {
        //Clean up our client list.
        SendMessage(pAV->hListC, LB_DELETESTRING, iConv, 0L);
        }
    else
        {
        ODSsz("Could not disconnect from %s", (LPSTR)szPeer);
        }

    ODS("TerminateConnection Confirmed.");
    return;
    }





/*
 * TerminateAllConnections
 *
 * Purpose:
 *  Walks our list of current connections and closes each.
 *
 * Parameters:
 *  pAV             PAPPVARS to the application variables.
 *
 * Return Value:
 *  None
 */

void TerminateAllConnections(PAPPVARS pAV)
    {
    int         i;
    int         cConv;

    ODS("TerminateAllConnections Entry.");

    cConv=(int)SendMessage(pAV->hListC, LB_GETCOUNT, 0, 0L);

    /*
     * We use zero as the index of the conversation since TerminateConnection
     * will delete the string from hListC.  If we tried passing indicies,
     * then the first one will be right and the rest wrong.  We just
     * delete each conversation in turn as they come to index 0.
     */
    for (i=0; i < cConv; i++)
        TerminateConnection(pAV, 0);


    D(cConv=(int)SendMessage(pAV->hListC, LB_GETCOUNT, 0, 0L));
    D(if (0!=cConv) ODS("TerminateAllConnections:  Still entries in hListC!"));

    ODS("TerminateAllConnections Confirmed.");
    return;
    }
