/*
 * INIT.C
 *
 * Initialization code for MazeWar.
 *
 * Copyright (c)1992 Microsoft Corporation, All Right Reserved
 *
 * Kraig Brockschmidt, Software Design Engineer
 * Microsoft Systems Developer Relations
 * One Microsoft Way
 * Redmond, WA  98052
 *
 * Internet  :  kraigb@microsoft.com
 * Compuserve:  70750,2344
 * Fax       :  (206)936-7329
 */

#define STRICT
#include <windows.h>
#include <nddeapi.h>
#include <ddeml.h>
#include <memory.h>
#include "exchange.h"




/*
 * AppPAllocate
 *
 * Purpose:
 *  Allocates and initializes an application variable data structure,
 *  registers window classes, and creates the main window.
 *
 * Parameters:
 *  pfSuccess       PINT in which a flag indicates the outcome of the function.
 *
 *  Other parameters are from WinMain
 *
 * Return Value:
 *  PAPPVARS        If NULL returned then AppPAllocate could not allocate
 *                  memory.  If a non-NULL pointer is returned with
 *                  *pfSuccess, then call AppPFree immediately.  If you
 *                  get a non-NULL pointer and *pfSuccess==TRUE then the
 *                  function succeeded.
 */

PAPPVARS AppPAllocate(PINT pfSuccess, HINSTANCE hInst, HINSTANCE hInstPrev
                      , LPSTR pszCmdLine, int nCmdShow)
    {
    PAPPVARS    pAV;
    WNDCLASS    wc;
    DWORD       dwVisible=0L;

    if (NULL==pfSuccess)
        return NULL;

    *pfSuccess=FALSE;

    //Allocate the structure
    pAV=(PAPPVARS)LocalAlloc(LPTR, CBAPPVARS);

    if (NULL==pAV)
        return NULL;

    //Initialize WinMain parameter holders.
    pAV->hInst=hInst;
    pAV->hInstPrev=hInstPrev;
    pAV->pszCmdLine=pszCmdLine;
    pAV->nCmdShow=nCmdShow;

    //Cache the strings from the stringtable, making the PSZ macro work.
    pAV->hStrings=AppHStringCache(hInst, pAV->rgpsz);

    if (NULL==pAV->hStrings)
        return pAV;


    //Go do DDEML initialization
    if (!AppDDEInitialize(pAV))
        return pAV;

    //Default conversation complexity
    pAV->iDL=ONEWAY;

    //Copy initial data into the pAV->xData field.
    lstrcpy(pAV->xData.szText, PSZ(IDS_CAPTION));
    pAV->xData.dw=72388L;


    //Register our window classes.
    if (!hInstPrev)
        {
        wc.style          = CS_HREDRAW | CS_VREDRAW;
        wc.lpfnWndProc    = ExchangeWndProc;
        wc.cbClsExtra     = 0;
        wc.cbWndExtra     = sizeof(LONG);
        wc.hInstance      = hInst;
        wc.hIcon          = LoadIcon(hInst, "Icon");
        wc.hCursor        = LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground  = (HBRUSH)(COLOR_WINDOW + 1);
        wc.lpszMenuName   = MAKEINTRESOURCE(IDR_MENU);
        wc.lpszClassName  = PSZ(IDS_CLASSNAME);

        if (!RegisterClass(&wc))
            return pAV;
        }

    //Create the main window.
    pAV->hWnd=CreateWindow(PSZ(IDS_CLASSNAME), PSZ(IDS_CAPTION)
                           , WS_MINIMIZEBOX | WS_OVERLAPPEDWINDOW
                           ,35, 35, 350, 240, NULL, NULL, hInst, pAV);


    if (NULL==pAV->hWnd)
        return pAV;

    /*
     * Create two listboxes in which to store peer names and conversation
     * handles.  PeerListClient are those that we DdeConnected to.
     * PeerListServer are those that connected to us.
     *
     * For sample code purposes and debugging, it makes a lot of sense
     * to show these lists because you can then visually see what
     * connections are around.  For non-debug, we take off the WS_VISIBLE
     * bit and keep things cleaner.
     */
    D(dwVisible=WS_VISIBLE);

    pAV->hListC=CreateWindow("listbox", "PeerListClient"
                             , dwVisible | WS_CAPTION | WS_CHILD | LBS_SORT
                             , 10, 80, 100, 100
                             , pAV->hWnd, (HMENU)1000, hInst, NULL);


    if (NULL==pAV->hListC)
        return pAV;

    pAV->hListS=CreateWindow("listbox", "PeerListServer"
                             , dwVisible | WS_CAPTION | WS_CHILD, 120, 80
                             , 100, 100, pAV->hWnd, (HMENU)1001, hInst, NULL);

    if (NULL==pAV->hListS)
        return pAV;


    ShowWindow(pAV->hWnd, nCmdShow);
    UpdateWindow(pAV->hWnd);

    *pfSuccess=TRUE;
    return pAV;
    }




/*
 * AppDDEInitialize
 *
 * Purpose:
 *  Performs all the DDEML-specific initialization for AppPAllocate
 *
 * Parameters:
 *  pAV             PAPPVARS to the application variables.
 *
 * Return Value:
 *  BOOL            TRUE if all initialization succeeded.  FALSE otherwise.
 */

BOOL AppDDEInitialize(PAPPVARS pAV)
    {
    UINT                    uRet;
    NDDESHAREINFO         ndsi;
    NDDESHAREITEMINFO     ndsii[2];

    //Attempt to initialize DDEML
    pAV->pfnDDEHandler=MakeProcInstance((FARPROC)DDETransactionHandler, pAV->hInst);

    if (NULL==pAV->pfnDDEHandler)
        return FALSE;

    /*
     * We're a standard client/server app that takes no executes.  Note
     * that we don't use CBF_FAIL_SELFCONNECTS because having Network DDE
     * in the picture means that DDEML itself won't detect self-connects.
     */
    uRet=DdeInitialize((LPDWORD)&pAV->dwDDE, (PFNCALLBACK)pAV->pfnDDEHandler
                       , APPCLASS_STANDARD | CBF_FAIL_EXECUTES
                         | CBF_SKIP_REGISTRATIONS | CBF_SKIP_UNREGISTRATIONS
                       , 0L);

    if (DMLERR_NO_ERROR!=uRet)
        return FALSE;

    /*
     * DDEML is cool, so we need to create all our string handles we'll
     * use in DDE conversations, then register the service name.
     */
    pAV->hszShare    =DdeCreateStringHandle(pAV->dwDDE, PSZ(IDS_SHARE),         CP_WINANSI);
    pAV->hszService  =DdeCreateStringHandle(pAV->dwDDE, PSZ(IDS_SERVICE),       CP_WINANSI);
    pAV->hszTopic    =DdeCreateStringHandle(pAV->dwDDE, PSZ(IDS_TOPIC),         CP_WINANSI);
    pAV->hszItemData =DdeCreateStringHandle(pAV->dwDDE, PSZ(IDS_ITEMDATA),      CP_WINANSI);
    pAV->hszItemLocal=DdeCreateStringHandle(pAV->dwDDE, PSZ(IDS_ITEMLOCALNAME), CP_WINANSI);
    pAV->hszItemPeer =DdeCreateStringHandle(pAV->dwDDE, PSZ(IDS_ITEMPEERLIST),  CP_WINANSI);

    if (NULL==pAV->hszShare || NULL==pAV->hszService
        || NULL==pAV->hszTopic || NULL==pAV->hszItemData
        || NULL==pAV->hszItemLocal || NULL==pAV->hszItemPeer)
        return FALSE;

    if (0==DdeNameService(pAV->dwDDE, pAV->hszService, (HSZ)NULL, DNS_REGISTER))
        return FALSE;

    //Define formats for the DDE data items we support.
    pAV->cfItem=RegisterClipboardFormat(PSZ(IDS_ITEMDATA));

    if (NULL==pAV->cfItem)
        return FALSE;


    /*
     * Add the services we support to the share list.
     * The first thing we do is delete our share if it exists.  We might
     * have changed our share since the last time we ran, so by deleting
     * it first we guarantee that the new information will be written.
     */

    //Delete the old share
    NDdeShareDel(NULL, PSZ(IDS_SHARE), 0);

    /*
     * Set up the password, application, and topic information.  NOTE: MEGA
     * IMPORTANT:  If you use passwords, store then in ALL UPPERCASE or
     * they will not match with a password entered through the password
     * dialog because that dialog uppercases everything.
     */
    memset(&ndsi, 0, sizeof(NDDESHAREINFO));
    lstrcpy(ndsi.szShareName, PSZ(IDS_SHARE));

    ndsi.lpszTargetApp  =PSZ(IDS_CLASSNAME);
    ndsi.lpszTargetTopic=PSZ(IDS_TOPIC);
    ndsi.lpszItem       =PSZ(IDS_ITEMDATA);
    ndsi.lpbPassword1   =(LPBYTE)"FULL";        //Be sure to UPPERCASE!
    ndsi.cbPassword1    =4;
    ndsi.lpbPassword2   =(LPBYTE)"NOSTART";     //Be sure to UPPERCASE!
    ndsi.cbPassword2    =7;

    ndsi.dwPermissions1 =NDDEACCESS_EXCLUDEEXECUTE;
    ndsi.dwPermissions2 =NDDEACCESS_REQUEST | NDDEACCESS_POKE | NDDEACCESS_ADVISE;

    ndsi.cAddItems=2;
    ndsi.lpNDdeShareItemInfo=ndsii;

    //Set up the additional items in this share.
    memset(ndsii, 0, 2*sizeof(NDDESHAREITEMINFO));
    ndsii[0].lpszItem=PSZ(IDS_ITEMPEERLIST);
    ndsii[0].dwPermissions=NDDEACCESS_REQUEST;      //We only request this item

    ndsii[1].lpszItem=PSZ(IDS_ITEMLOCALNAME);
    ndsii[1].dwPermissions=NDDEACCESS_POKE;         //We only poke this item

    //Go add the share
    uRet=NDdeShareAdd(NULL, 2, &ndsi, sizeof(ndsi));

    if (NDDE_NO_ERROR!=uRet)
        return FALSE;

    //Retrieve the local name of this machine.
    uRet=NDdeGetNodeName(pAV->szLocal, sizeof(pAV->szLocal));
    return (NDDE_NO_ERROR==uRet);
    }





/*
 * AppPFree
 *
 * Purpose:
 *  Reverses all initialization done by AppPAllocate, cleaning up
 *  any allocations including the application structure itself.
 *
 * Parameters:
 *  pAV             PAPPVARS to the application structure
 *
 * Return Value:
 *  PAPPVARS        NULL if successful, pAV if not, meaning we couldn't
 *                  free some allocation.
 */

PAPPVARS AppPFree(PAPPVARS pAV)
    {
    if (NULL==pAV)
        return NULL;

    if (IsWindow(pAV->hListS))
        DestroyWindow(pAV->hListS);

    if (IsWindow(pAV->hListC))
        DestroyWindow(pAV->hListC);

    if (IsWindow(pAV->hWnd))
        DestroyWindow(pAV->hWnd);

    AppDDEFree(pAV);

    if (NULL!=pAV->hStrings)
        LocalFree(pAV->hStrings);

    return (PAPPVARS)LocalFree((HLOCAL)pAV);
    }




/*
 * AppDDEFree
 *
 * Purpose:
 *  Performs all the DDEML-specific cleanup for AppPFree
 *
 * Parameters:
 *  pAV             PAPPVARS to the application variables.
 *
 * Return Value:
 *  BOOL            TRUE if all cleanup succeeded.  FALSE otherwise.
 */

BOOL AppDDEFree(PAPPVARS pAV)
    {
    if (NULL!=pAV->hszItemPeer)
        DdeFreeStringHandle(pAV->dwDDE, pAV->hszItemPeer);

    if (NULL!=pAV->hszItemLocal)
        DdeFreeStringHandle(pAV->dwDDE, pAV->hszItemLocal);

    if (NULL!=pAV->hszItemData)
        DdeFreeStringHandle(pAV->dwDDE, pAV->hszItemData);

    if (NULL!=pAV->hszTopic)
        DdeFreeStringHandle(pAV->dwDDE, pAV->hszTopic);

    if (NULL!=pAV->hszService)
        {
        DdeNameService(pAV->dwDDE, pAV->hszService, NULL, DNS_UNREGISTER);
        DdeFreeStringHandle(pAV->dwDDE, pAV->hszService);
        }

    if (NULL!=pAV->hszShare)
        DdeFreeStringHandle(pAV->dwDDE, pAV->hszShare);

    if (NULL!=pAV->dwDDE)
        DdeUninitialize(pAV->dwDDE);

    return TRUE;
    }







/*
 * AppHStringCache
 *
 * Purpose:
 *  Allocates FIXED local memory and reads the applications string
 *  resources into that memory.  Pointers to the strings in the memory
 *  are then stored at ppsz which is assumed to be CSTRINGS long.
 *
 * Parameters:
 *  hInst           HINSTANCE of the application
 *  ppsz            char ** to an array in which to store pointers.
 *
 * Return Value:
 *  HANDLE          Handle to the local memory.  NULL if memory could
 *                  not be allocated.
 */

HANDLE AppHStringCache(HINSTANCE hInst, char **ppsz)
    {
    HANDLE      hMem;
    char       *pch;
    UINT        cchUsed=0;
    UINT        cch;
    short       i;

    hMem=LocalAlloc(LPTR, CSTRINGS*CCHSTRINGMAX);

    if (NULL!=hMem)
        {
        //Only valid for local fixed memory.
        pch=(char *)hMem;

        /*
         * Load the strings into the memory and retain the specific
         * pointer to that string.
         */
        for (i=0; i<CSTRINGS; i++)
            {
            cch=LoadString(hInst, i, (LPSTR)(pch+cchUsed), CCHSTRINGMAX-1);
            ppsz[i]=(char *)(pch+cchUsed);

            /*
             * Add one to cch to include a NULL.  The memory was ZEROINITed
             * on allocation so by skipping a byte we get the NULL.
             */
            cchUsed +=cch+1;
            }

        /*
         * We are assuming that no string is over CCHSTRINGMAX, and therefore
         * we did not use all the allocated memory.  Therefore LocalReAlloc
         * will only SHRINK the block, never expand it.  So if it fails, we
         * don't care--all the strings are still there, we just wasted some
         * space.
         */
        LocalReAlloc(hMem, cchUsed+1, LPTR);
        }

    return hMem;
    }
